@@ -65,6 +65,7 @@ pm_path_compare(struct f_path_mask *m1, struct f_path_mask *m2)
     if ((!m1) || (!m2))
       return !((!m1) && (!m2));
 
+    /* FIXME: buggy, should return -1, 0, 1; but it doesn't matter */
     if ((m1->kind != m2->kind) || (m1->val != m2->val)) return 1;
     m1 = m1->next;
     m2 = m2->next;
@@ -111,6 +112,13 @@ pm_format(struct f_path_mask *p, byte *buf, unsigned int size)
   *buf = 0;
 }
 
+static inline int int_cmp(int i1, int i2)
+{
+  if (i1 == i2) return 0;
+  if (i1 < i2) return -1;
+  else return 1;
+}
+
 /**
  * val_compare - compare two values
  * @v1: first value
@@ -133,6 +141,14 @@ val_compare(struct f_val v1, struct f_val v2)
     return 1;
 
   if (v1.type != v2.type) {
+#ifndef IPV6
+    /* IP->Quad implicit conversion */
+    if ((v1.type == T_QUAD) && (v2.type == T_IP))
+      return int_cmp(v1.val.i, ipa_to_u32(v2.val.px.ip));
+    if ((v1.type == T_IP) && (v2.type == T_QUAD))
+      return int_cmp(ipa_to_u32(v1.val.px.ip), v2.val.i);
+#endif
+
     debug( "Types do not match in val_compare\n" );
     return CMP_ERROR;
   }
@@ -141,9 +157,8 @@ val_compare(struct f_val v1, struct f_val v2)
   case T_INT:
   case T_BOOL:
   case T_PAIR:
-    if (v1.val.i == v2.val.i) return 0;
-    if (v1.val.i < v2.val.i) return -1;
-    return 1;
+  case T_QUAD:
+    return int_cmp(v1.val.i, v2.val.i);
   case T_IP:
     return ipa_compare(v1.val.px.ip, v2.val.px.ip);
   case T_PREFIX:
@@ -164,6 +179,11 @@ val_compare(struct f_val v1, struct f_val v2)
   }
 }
 
+int 
+tree_compare(const void *p1, const void *p2)
+{
+  return val_compare((* (struct f_tree **) p1)->from, (* (struct f_tree **) p2)->from);
+}
 
 void
 f_prefix_get_bounds(struct f_prefix *px, int *l, int *h)
@@ -191,8 +211,13 @@ val_simple_in_range(struct f_val v1, struct f_val v2)
 {
   if ((v1.type == T_PATH) && (v2.type == T_PATH_MASK))
     return as_path_match(v1.val.ad, v2.val.path_mask);
-  if ((v1.type == T_PAIR) && (v2.type == T_CLIST))
+  if (((v1.type == T_PAIR) || (v1.type == T_QUAD)) && (v2.type == T_CLIST))
     return int_set_contains(v2.val.ad, v1.val.i);
+#ifndef IPV6
+  /* IP->Quad implicit conversion */
+  if ((v1.type == T_IP) && (v2.type == T_CLIST))
+    return int_set_contains(v2.val.ad, ipa_to_u32(v1.val.px.ip));
+#endif
   if ((v1.type == T_STRING) && (v2.type == T_STRING))
     return patmatch(v2.val.s, v1.val.s);
 
@@ -205,6 +230,76 @@ val_simple_in_range(struct f_val v1, struct f_val v2)
   return CMP_ERROR;
 }
 
+static int
+clist_set_type(struct f_tree *set, struct f_val *v)
+{
+ switch (set->from.type) {
+  case T_PAIR:
+    v->type = T_PAIR;
+    return 1;
+  case T_QUAD:
+#ifndef IPV6
+  case T_IP:
+#endif
+    v->type = T_QUAD;
+    return 1;
+    break;
+  default:
+    v->type = T_VOID;
+    return 0;
+  }
+}
+
+static int
+clist_match_set(struct adata *clist, struct f_tree *set)
+{
+  if (!clist)
+    return 0;
+
+  struct f_val v;
+  if (!clist_set_type(set, &v))
+    return CMP_ERROR;
+
+  u32 *l = (u32 *) clist->data;
+  u32 *end = l + clist->length/4;
+  while (l < end) {
+    v.val.i = *l++;
+    if (find_tree(set, v))
+      return 1;
+  }
+  return 0;
+}
+
+static struct adata *
+clist_del_matching(struct linpool *pool, struct adata *clist, struct f_tree *set)
+{
+  if (!clist)
+    return NULL;
+
+  struct f_val v;
+  clist_set_type(set, &v);
+
+  u32 tmp[clist->length/4];
+  u32 *l = (u32 *) clist->data;
+  u32 *k = tmp;
+  u32 *end = l + clist->length/4;
+
+  while (l < end) {
+    v.val.i = *l++;
+    if (!find_tree(set, v))
+      *k++ = v.val.i;
+  }
+
+  int nl = (k - tmp) * 4;
+  if (nl == clist->length)
+    return clist;
+
+  struct adata *res = lp_alloc(pool, sizeof(struct adata) + nl);
+  res->length = nl;
+  memcpy(res->data, tmp, nl);
+  return res;
+}
+
 /**
  * val_in_range - implement |~| operator
  * @v1: element
@@ -226,12 +321,16 @@ val_in_range(struct f_val v1, struct f_val v2)
   if ((v1.type == T_PREFIX) && (v2.type == T_PREFIX_SET))
     return trie_match_prefix(v2.val.ti, &v1.val.px);
 
+  if ((v1.type == T_CLIST) && (v2.type == T_SET))
+    return clist_match_set(v1.val.ad, v2.val.t);
+
   if (v2.type == T_SET)
     switch (v1.type) {
     case T_ENUM:
     case T_INT:
+    case T_PAIR:
+    case T_QUAD:
     case T_IP:
-    case T_PREFIX:
       {
 	struct f_tree *n;
 	n = find_tree(v2.val.t, v1);
@@ -275,6 +374,7 @@ val_print(struct f_val v)
   case T_IP: PRINTF( "%I", v.val.px.ip ); break;
   case T_PREFIX: PRINTF( "%I/%d", v.val.px.ip, v.val.px.len ); break;
   case T_PAIR: PRINTF( "(%d,%d)", v.val.i >> 16, v.val.i & 0xffff ); break;
+  case T_QUAD: PRINTF( "%R", v.val.i ); break;
   case T_PREFIX_SET: trie_print(v.val.ti, buf, 2040); break;
   case T_SET: tree_print( v.val.t ); PRINTF( "\n" ); break;
   case T_ENUM: PRINTF( "(enum %x)%d", v.type, v.val.i ); break;
@@ -350,7 +450,7 @@ static struct f_val
 interpret(struct f_inst *what)
 {
   struct symbol *sym;
-  struct f_val v1, v2, res;
+  struct f_val v1, v2, res, *vp;
   unsigned u1, u2;
   int i;
   u32 as;
@@ -430,11 +530,11 @@ interpret(struct f_inst *what)
 /* Relational operators */
 
 #define COMPARE(x) \
-    TWOARGS_C; \
-    res.type = T_BOOL; \
+    TWOARGS; \
     i = val_compare(v1, v2); \
     if (i==CMP_ERROR) \
-      runtime( "Error in comparison" ); \
+      runtime( "Can't compare values of incompatible types" ); \
+    res.type = T_BOOL; \
     res.val.i = (x); \
     break;
 
@@ -468,27 +568,19 @@ interpret(struct f_inst *what)
   case 's':
     ARG(v2, a2.p);
     sym = what->a1.p;
-    switch (res.type = v2.type) {
-    case T_VOID: runtime( "Can't assign void values" );
-    case T_ENUM:
-    case T_BOOL:
-    case T_INT:
-    case T_PAIR:
-    case T_STRING:
-    case T_IP:
-    case T_PREFIX:
-    case T_PREFIX_SET:
-    case T_SET:
-    case T_PATH:
-    case T_PATH_MASK:
-    case T_CLIST:
-      if (sym->class != (SYM_VARIABLE | v2.type))
-	runtime( "Assigning to variable of incompatible type" );
-      * (struct f_val *) sym->def = v2; 
-      break;
-    default:
-      bug( "Set to invalid type" );
+    vp = sym->def;
+    if ((sym->class != (SYM_VARIABLE | v2.type)) && (v2.type != T_VOID)) {
+#ifndef IPV6
+      /* IP->Quad implicit conversion */
+      if ((sym->class == (SYM_VARIABLE | T_QUAD)) && (v2.type == T_IP)) {
+	vp->type = T_QUAD;
+	vp->val.i = ipa_to_u32(v2.val.px.ip);
+	break;
+      }
+#endif
+      runtime( "Assigning to variable of incompatible type" );
     }
+    *vp = v2; 
     break;
 
     /* some constants have value in a2, some in *a1.p, strange. */
@@ -600,42 +692,48 @@ interpret(struct f_inst *what)
 	e = ea_find( (*f_tmp_attrs), what->a2.i );
       if ((!e) && (f_flags & FF_FORCE_TMPATTR))
 	e = ea_find( (*f_rte)->attrs->eattrs, what->a2.i );
-      
-      switch (what->aux & EAF_TYPE_MASK) {
-      case EAF_TYPE_INT:
-	if (!e) {
-	  res.type = T_VOID;
+
+      if (!e) {
+	/* A special case: undefined int_set looks like empty int_set */
+	if ((what->aux & EAF_TYPE_MASK) == EAF_TYPE_INT_SET) {
+	  res.type = T_CLIST;
+	  res.val.ad = adata_empty(f_pool);
 	  break;
 	}
+	/* Undefined value */
+	res.type = T_VOID;
+	break;
+      }
+
+      switch (what->aux & EAF_TYPE_MASK) {
+      case EAF_TYPE_INT:
 	res.type = T_INT;
 	res.val.i = e->u.data;
 	break;
+      case EAF_TYPE_ROUTER_ID:
+	res.type = T_QUAD;
+	res.val.i = e->u.data;
+	break;
+      case EAF_TYPE_OPAQUE:
+	res.type = T_ENUM_EMPTY;
+	res.val.i = 0;
+	break;
       case EAF_TYPE_IP_ADDRESS:
-	if (!e) {
-	  res.type = T_VOID;
-	  break;
-	}
 	res.type = T_IP;
 	struct adata * ad = e->u.ptr;
 	res.val.px.ip = * (ip_addr *) ad->data;
 	break;
       case EAF_TYPE_AS_PATH:
-	if (!e) {
-	  res.type = T_VOID;
-	  break;
-	}
         res.type = T_PATH;
 	res.val.ad = e->u.ptr;
 	break;
       case EAF_TYPE_INT_SET:
-	if (!e) {
-	  res.type = T_CLIST;
-	  res.val.ad = adata_empty(f_pool);
-	  break;
-	}
 	res.type = T_CLIST;
 	res.val.ad = e->u.ptr;
 	break;
+      case EAF_TYPE_UNDEF:
+	res.type = T_VOID;
+	break;
       default:
 	bug("Unknown type in e,a");
       }
@@ -654,10 +752,14 @@ interpret(struct f_inst *what)
       l->attrs[0].type = what->aux | EAF_ORIGINATED;
       switch (what->aux & EAF_TYPE_MASK) {
       case EAF_TYPE_INT:
+      case EAF_TYPE_ROUTER_ID:
 	if (v1.type != T_INT)
 	  runtime( "Setting int attribute to non-int value" );
 	l->attrs[0].u.data = v1.val.i;
 	break;
+      case EAF_TYPE_OPAQUE:
+	runtime( "Setting opaque attribute is not allowed" );
+	break;
       case EAF_TYPE_IP_ADDRESS:
 	if (v1.type != T_IP)
 	  runtime( "Setting ip attribute to non-ip value" );
@@ -731,7 +833,7 @@ interpret(struct f_inst *what)
   case P('a','f'):	/* Get first ASN from AS PATH */
     ONEARG;
     if (v1.type != T_PATH)
-      runtime( "AS Path expected" );
+      runtime( "AS path expected" );
 
     as = 0;
     as_path_get_first(v1.val.ad, &as);
@@ -760,6 +862,10 @@ interpret(struct f_inst *what)
       return res;
     res.type &= ~T_RETURN;    
     break;
+  case P('c','v'):	/* Clear local variables */
+    for (sym = what->a1.p; sym != NULL; sym = sym->aux2)
+      ((struct f_val *) sym->def)->type = T_VOID;
+    break;
   case P('S','W'):
     ONEARG;
     {
@@ -811,13 +917,25 @@ interpret(struct f_inst *what)
     TWOARGS;
     if (v1.type != T_CLIST)
       runtime("Can't add/delete to non-clist");
-    if (v2.type != T_PAIR)
+
+    struct f_val dummy;
+    if ((v2.type == T_PAIR) || (v2.type == T_QUAD))
+      i = v2.val.i;
+#ifndef IPV6
+    /* IP->Quad implicit conversion */
+    else if (v2.type == T_IP)
+      i = ipa_to_u32(v2.val.px.ip);
+#endif
+    else if ((v2.type == T_SET) && (what->aux == 'd') && clist_set_type(v2.val.t, &dummy))
+      what->aux = 'D';
+    else
       runtime("Can't add/delete non-pair");
 
     res.type = T_CLIST;
     switch (what->aux) {
-    case 'a': res.val.ad = int_set_add(f_pool, v1.val.ad, v2.val.i); break;
-    case 'd': res.val.ad = int_set_del(f_pool, v1.val.ad, v2.val.i); break;
+    case 'a': res.val.ad = int_set_add(f_pool, v1.val.ad, i); break;
+    case 'd': res.val.ad = int_set_del(f_pool, v1.val.ad, i); break;
+    case 'D': res.val.ad = clist_del_matching(f_pool, v1.val.ad, v2.val.t); break;
     default: bug("unknown Ca operation");
     }
     break;
@@ -938,10 +1056,13 @@ i_same(struct f_inst *f1, struct f_inst *f2)
 	       return 0; 
 	     f2->a2.p = f1->a2.p;
 	     break;
+  case P('c','v'): break; /* internal instruction */ 
   case P('S','W'): ONEARG; if (!same_tree(f1->a2.p, f2->a2.p)) return 0; break;
   case P('i','M'): TWOARGS; break;
   case P('A','p'): TWOARGS; break;
   case P('C','a'): TWOARGS; break;
+  case P('a','f'):
+  case P('a','l'): ONEARG; break;
   default:
     bug( "Unknown instruction %d in same (%c)", f1->code, f1->code & 0xff);
   }
