@@ -9,6 +9,8 @@
 #ifndef _BIRD_ATTRS_H_
 #define _BIRD_ATTRS_H_
 
+#include <stdint.h>
+
 /* a-path.c */
 
 #define AS_PATH_SET		1	/* Types of path segments */
@@ -28,6 +30,7 @@ int as_path_convert_to_old(struct adata *path, byte *dst, int *new_used);
 int as_path_convert_to_new(struct adata *path, byte *dst, int req_as);
 void as_path_format(struct adata *path, byte *buf, unsigned int size);
 int as_path_getlen(struct adata *path);
+int as_path_getlen_int(struct adata *path, int bs);
 int as_path_get_first(struct adata *path, u32 *orig_as);
 int as_path_get_last(struct adata *path, u32 *last_as);
 int as_path_is_member(struct adata *path, u32 as);
@@ -35,11 +38,12 @@ int as_path_is_member(struct adata *path, u32 as);
 #define PM_ASN		0
 #define PM_QUESTION	1
 #define PM_ASTERISK	2
+#define PM_ASN_EXPR	3
 
 struct f_path_mask {
   struct f_path_mask *next;
   int kind;
-  u32 val;
+  uintptr_t val;
 };
 
 int as_path_match(struct adata *path, struct f_path_mask *mask);
