@@ -43,10 +43,11 @@
 #include "lib/string.h"
 #include "lib/alloca.h"
 
+pool *rt_table_pool;
+
 static slab *rte_slab;
 static linpool *rte_update_pool;
 
-static pool *rt_table_pool;
 static list routing_tables;
 
 static void rt_format_via(rte *e, byte *via);
@@ -342,7 +343,7 @@ rte_validate(rte *e)
   int c;
   net *n = e->net;
 
-  if (ipa_nonzero(ipa_and(n->n.prefix, ipa_not(ipa_mkmask(n->n.pxlen)))))
+  if ((n->n.pxlen > BITS_PER_IP_ADDRESS) || !ip_is_prefix(n->n.prefix,n->n.pxlen))
     {
       log(L_BUG "Ignoring bogus prefix %I/%d received via %s",
 	  n->n.prefix, n->n.pxlen, e->sender->name);
