@@ -11,7 +11,7 @@
 #define _BIRD_OSPF_LSREQ_H_
 
 void ospf_lsreq_send(struct ospf_neighbor *n);
-void ospf_lsreq_receive(struct ospf_lsreq_packet *ps,
-			struct ospf_iface *ifa, struct ospf_neighbor *n);
+void ospf_lsreq_receive(struct ospf_packet *ps_i, struct ospf_iface *ifa,
+			struct ospf_neighbor *n);
 
 #endif /* _BIRD_OSPF_LSREQ_H_ */
