@@ -684,7 +684,7 @@ new_iface(struct proto *p, struct iface *new, unsigned long flags, struct iface_
     DBG( "Doing multicasts!\n" );
 
   rif->sock = sk_new( p->pool );
-  rif->sock->type = rif->multicast?SK_UDP_MC:SK_UDP;
+  rif->sock->type = SK_UDP;
   rif->sock->sport = P_CF->port;
   rif->sock->rx_hook = rip_rx;
   rif->sock->data = rif;
@@ -721,19 +721,38 @@ new_iface(struct proto *p, struct iface *new, unsigned long flags, struct iface_
   if (!ipa_nonzero(rif->sock->daddr)) {
     if (rif->iface)
       log( L_WARN "%s: interface %s is too strange for me", p->name, rif->iface->name );
-  } else if (sk_open(rif->sock)<0) {
-    log( L_ERR "%s: could not create socket for %s", p->name, rif->iface ? rif->iface->name : "(dummy)" );
-    if (rif->iface) {
-      rfree(rif->sock);
-      mb_free(rif);
-      return NULL;
-    }
-    /* On dummy, we just return non-working socket, so that user gets error every time anyone requests table */
+  } else {
+
+    if (sk_open(rif->sock)<0)
+      goto err;
+
+    if (rif->multicast)
+      {
+	if (sk_setup_multicast(rif->sock) < 0)
+	  goto err;
+	if (sk_join_group(rif->sock, rif->sock->daddr) < 0)
+	  goto err;
+      }
+    else
+      {
+	if (sk_set_broadcast(rif->sock, 1) < 0)
+	  goto err;
+      }
   }
 
   TRACE(D_EVENTS, "Listening on %s, port %d, mode %s (%I)", rif->iface ? rif->iface->name : "(dummy)", P_CF->port, rif->multicast ? "multicast" : "broadcast", rif->sock->daddr );
   
   return rif;
+
+ err:
+  log( L_ERR "%s: could not create socket for %s", p->name, rif->iface ? rif->iface->name : "(dummy)" );
+  if (rif->iface) {
+    rfree(rif->sock);
+    mb_free(rif);
+    return NULL;
+  }
+  /* On dummy, we just return non-working socket, so that user gets error every time anyone requests table */
+  return rif;
 }
 
 static void
@@ -742,7 +761,7 @@ rip_real_if_add(struct object_lock *lock)
   struct iface *iface = lock->iface;
   struct proto *p = lock->data;
   struct rip_interface *rif;
-  struct iface_patt *k = iface_patt_find(&P_CF->iface_list, iface);
+  struct iface_patt *k = iface_patt_find(&P_CF->iface_list, iface, iface->addr);
 
   if (!k)
     bug("This can not happen! It existed few seconds ago!" );
@@ -771,7 +790,7 @@ rip_if_notify(struct proto *p, unsigned c, struct iface *iface)
     }
   }
   if (c & IF_CHANGE_UP) {
-    struct iface_patt *k = iface_patt_find(&P_CF->iface_list, iface);
+    struct iface_patt *k = iface_patt_find(&P_CF->iface_list, iface, iface->addr);
     struct object_lock *lock;
     struct rip_patt *PATT = (struct rip_patt *) k;
 
@@ -846,7 +865,8 @@ rip_store_tmp_attrs(struct rte *rt, struct ea_list *attrs)
  * own), so store it into our data structures. 
  */
 static void
-rip_rt_notify(struct proto *p, struct network *net, struct rte *new, struct rte *old, struct ea_list *attrs)
+rip_rt_notify(struct proto *p, struct rtable *table UNUSED, struct network *net,
+	      struct rte *new, struct rte *old, struct ea_list *attrs)
 {
   CHK_MAGIC;
 
@@ -937,7 +957,7 @@ rip_rte_insert(net *net UNUSED, rte *rte)
 static void
 rip_rte_remove(net *net UNUSED, rte *rte)
 {
-  struct proto *p = rte->attrs->proto;
+  // struct proto *p = rte->attrs->proto;
   CHK_MAGIC;
   DBG( "rip_rte_remove: %p\n", rte );
   rem_node( &rte->u.rip.garbage );
